/*
 * True and cryptographic random number generation.
 *
 * (c) Copyright 1990-1996 by Philip Zimmermann.  All rights reserved.
 * The author assumes no liability for damages resulting from the use
 * of this software, even if the damage results from defects in this
 * software.  No warranty is expressed or implied.
 *
 * Note that while most PGP source modules bear Philip Zimmermann's
 * copyright notice, many of them have been revised or entirely written
 * by contributors who frequently failed to put their names in their
 * code.  Code that has been incorporated into PGP from other authors
 * was either originally published in the public domain or is used with
 * permission from the various authors.
 *
 * PGP is available for free to the public under certain restrictions.
 * See the PGP User's Guide (included in the release package) for
 * important information about licensing, patent restrictions on
 * certain algorithms, trademarks, copyrights, and export controls.
 *
 * Written by Colin Plumb.
 * 
 * Modifications for Newt/slang interface
 * Copyright (c) 2001 Red Hat, Inc.
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <newt.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <signal.h>	/* For SIGINT */
#include <time.h>
#include <unistd.h>
#include <fcntl.h>

#include "noise.h" /* for noise() */
#include "../librand/rand.h" /* for truerand stuff. */

static unsigned trueRandBits = 0;	/* Bits of entropy in pool */

#define RANDPOOLBITS 3079

/*
 * Given an event (typically a keystroke) coded by "event"
 * at a random time, add all randomness to the random pool,
 * compute a (conservative) estimate of the amount, add it
 * to the pool, and return the amount of randomness.
 * (The return value is just for informational purposes.)
 *
 * Double events are okay, but three in a row is considered
 * suspiscous and the randomness is counted as 0.
 */
unsigned
trueRandEvent(int event)
{
	static int event1 = 0, event2 = 0;
	word32 delta;
	unsigned cbits;

	delta = noise();
	randPoolAddBytes((byte *)&event, sizeof(event));

	if (event == event1 && event == event2) {
		cbits = 0;
	} else {
		event2 = event1;
		event1 = event;

		for (cbits = 0; delta; cbits++)
			delta >>= 1;

		/* Excessive paranoia? */

		if (cbits > 8)
			cbits = 8;
	}

	trueRandBits += cbits;
	if (trueRandBits > RANDPOOLBITS)
		trueRandBits = RANDPOOLBITS;

	return cbits;
}

/*
 * Performs an accumulation of random bits.  As long as there are fewer bits
 * in the buffer than are needed (the number passed, plus pending bits),
 * prompt for more.
 */
void
trueRandAccum(unsigned count)	/* Get this many random bits ready */
{
    int c;
    newtComponent scale;
    newtComponent title;
    newtComponent form;    
    char message[1024];
    newtGrid box;
    
    if (count > RANDPOOLBITS)
        count = RANDPOOLBITS;
    
    if (trueRandBits >= count)
        return;

    box = newtCreateGrid(1, 3);

    sprintf(message,"We need to generate %u random bits.  This "
	    "is done by measuring the time intervals "
	    "between your keystrokes.  Please enter some"
	    " random text on your keyboard.", count);

    title = newtTextboxReflowed(1, 1, message,45,10,0,0);
    
    newtGridSetField(box, 0, 0, NEWT_GRID_COMPONENT, title,
		     0, 0, 0, 0, 0, 0);
    
    /* The progress bar */
    scale = newtScale(0, 0, 30, count);
    newtScaleSet(scale, 0);
    
    newtGridSetField(box, 0, 1, NEWT_GRID_COMPONENT, scale,
		     0, 1, 0, 0, 0, 0);
    
    form = newtForm(NULL, NULL, 0);
    newtGridAddComponentsToForm(box, form, 1);
    
    newtGridWrappedWindow(box, "Collecting random data");
    
    newtDrawForm(form);
    
    do {
        newtScaleSet(scale, trueRandBits);
	newtRefresh();
	
	c = SLang_getkey();	/* always wait for input */
	SLang_flush_input();
	trueRandEvent(c);
    } while (trueRandBits < count);
    
    newtScaleSet(scale, count);
    newtRefresh();

    sleep(1);
    newtPopWindow();
    newtFormDestroy(form);
    newtClearKeyBuffer();
}

int main(int argc, char *argv[])
{
  char buf[1024];
  int n, rows, cols, fd;

  if (argc != 3) {
      fprintf(stderr, "Usage: keyrand <number-of-bits> <output-file>\n");
      exit(1);
  }
    
  newtGetScreenSize(&cols, &rows);

  newtInit();
  newtCls();

  newtDrawRootText(0, 0, 
		   "Red Hat Keypair Generation (c) 2001-2002 Red Hat, Inc.");

  newtDrawRootText(cols-32, rows-3, "This program contains software");
  newtDrawRootText(cols-32, rows-2, "written by Philip Zimmerman");

  n = atoi(argv[1]);
  fd = open(argv[2], O_WRONLY|O_APPEND);

  if (fd < 0) {
      newtWinMessage("Error", "Exit", "Random file not found");
      newtFinished();
      exit(1);
  }

  trueRandAccum(n);

  n = n/8; /* bytes */

  randPoolGetBytes(buf, n);

  if (write(fd, buf, n) != n || close(fd)) {
      newtWinMessage("Error", "Exit", "Error writing to random file");
      newtFinished();
      exit(1);
  }

  newtFinished();

  return 0;
}
